import mapValues from 'lodash/mapValues';
import fetch from 'isomorphic-fetch';
import httpErrors from 'http-errors';
import Route from './route';

fetch.Promise = Promise;

/**
 * @class BaseSdk
 * @param opts
 */
class BaseSdk {
  constructor(opts) {
    if (typeof opts === 'string') {
      opts = {url: opts};
    }

    const url = opts.url;

    if (!url) {
      throw new Error('Service base url must be specified');
    }

    this._baseUrl = ((url.indexOf('http') !== 0) ? `http://${url}` : url).replace(/\/+$/, '');
    this._opts = opts;
  }

  /**
   * Process http request.
   * @param {string} method The http method (get, post, put, delete, head, etc).
   * @param {string} url The service endpoint url.
   * @param {Object=} request The request object.
   * @returns {Promise<{response: Object, body: Object|[]}>}
   * @protected
   */
  fetch(method, url, request = {}) {
    const headers = Object.assign({
      'Accept': 'application/json',
      'Content-Type': 'application/json'
    }, request.headers || {});

    method = method.toLowerCase();

    const params = {
      method,
      headers,
      credentials: 'same-origin'
    };

    if ((method === 'post' || method === 'put') && request.body) {
      params.body = JSON.stringify(request.body);
    }

    return fetch(url, params)
      .catch(() => { throw httpErrors.InternalServerError('Server is unavailable'); })
      .then(this.assertResponse)
  }

  /**
   * Asser response
   * @param {Object=} response The response object.
   * @returns {Promise.<{response: Object, body: Object|[]}>}
   * @protected
   */
  assertResponse(response) {
    const status = response.status;

    if (status < 400) {
      return (status === 204) ? null : response.json();
    }

    return response.json().then(error => {
      throw httpErrors(status, error)
    });
  }

  /**
   * Resolves path to full service absolute url.
   * @param {string=} path The service path
   * @returns {string}
   */
  url(path) {
    return this._baseUrl.concat('/').concat((path || '').replace(/^\/+/, ''));
  }
}

/**
 * SDK class factory method. Creates a class where methods are generated by specified
 * route config object.
 * @param {object} sdkConfig The object where values are {@link Route} instances.
 * @returns {function}
 */
function sdkFactory(sdkConfig) {
  /**
   * Generated SDK class.
   * @constructor
   */
  class Sdk extends BaseSdk {}

  Sdk.errors = httpErrors;

  const routes = mapValues(sdkConfig, routeFactory);
  Object.assign(Sdk.prototype, mapValues(routes, routeMethod));

  return Sdk;
}

/**
 * Creates a route instance
 * @param routeConfig
 * @param name
 * @returns {Route}
 */
function routeFactory(routeConfig, name) {
  if (typeof routeConfig === 'string') {
    routeConfig = { path: routeConfig };
  }

  routeConfig.name = name;

  return new Route(routeConfig);
}

/**
 * Create request method for SDK library that corresponds to a specified route.
 * @param {Route} route
 * @returns {function(request:Object)}
 */
function routeMethod(route) {
  /**
   * Generated request method that corresponds to a specified route.
   * @param {{params:Object=, query:Object=, data:Object=, headers:Object=}=} request
   * @this {BaseSdk}
   * @returns {Promise.<Object|[]>}
   */
  return function (request) {
    request = request || {};
    const path = route.pathFormat(request);

    return this.fetch(route.method, this.url(path), request);
  };
}

export default sdkFactory;
